'use strict';

/*
 * Functions to help with a functional-style approach based on generators.
 *
 * I could have used something like lodash here, but for these sort of challenges, it's fun to write your own, and I got to play with generators.
 * 
 */

import { assertIsNotNull } from './assertions.js';

// A standard map() function, but for iterators / generators. Applies the callback to each item in the collection (which takes the current value and the value's index
// in the collection).

function map(collection, callback) {
	assertIsNotNull(collection);
	return asReiterable(function* () {
		let index = 0;
		for (let value of asIterable(collection)) {
			yield callback(value, index++);
		}
	});
}

// A standard reduce() function, but for iterators / generators. Folds the collection into a single value using the provided callback (which takes the accumulator and
// the current value).

function reduce(collection, callback, initial) {
	assertIsNotNull(collection);
	let accumulator = initial;
	for (let value of asIterable(collection)) {
		accumulator = callback(accumulator, value);
	}
	return accumulator;
}

// Called before iterating over any object to provide a single hook for adding functionality, such as the ability to iterator over name-value pairs of object members.

function asIterable(object) {
	assertIsNotNull(object);
	return Symbol.iterator in object ? object : iterateProperties(object);
}

// Allows simple objects to be iterated by returing a name-value tuple.

function iterateProperties(object) {
	assertIsNotNull(object);
	return asReiterable(function* () {
		for (let property in object) {
			if (object.hasOwnProperty(property)) {
				yield {
					name: property,
					value: object[property]
				};
			}
		}
	});
}

// Converts an iterable into an array so that we can manipulate the collection using methods available to arrays.

function toArray(iterable) {
	assertIsNotNull(iterable);
	return [ ...iterable ];
}

// We're passing iterators around in the code; this wrapper method allows us to express that we want to be able to fetch results from the iterator by index.

function asIndexable(iterable) {
	assertIsNotNull(iterable);
	return memoize(iterable);
}

// Iterators can only be iterated once. This wrapper method allows the calling code to express that it needs to be able to re-iterate the iterator generated by this generator.

function asReiterable(generator) {
	// This originally had a separate implementation that proved buggy, so fall back on memoize() for now.
	return memoize(generator());
}

// Iterators can only be iterated once. This wrapper method caches the results of the iterator as they are encoutered to support re-iteration and accessing results by index.
//
// We could have passed bound generators around rather than iterators, but while all generators are iterators, not all iterators are generators, and a decision had to be made
// one way or the other.
//
// Thie method also provides a single implementation of results caching for when it makes sense for the caller to decider whether or not caching is appropriate.

function memoize(sourceIterable) {
	assertIsNotNull(sourceIterable);

	// If we've already memoized the iterable, don't do so again.

	if (sourceIterable._memoization) {
		return sourceIterable._memoization;
	}

	// If we've been passed an array, then we can use it as the underlying cache; if not, we'll have to populate our own. (Following a functional style, we're assuming that
	// arrays passed in won't subsequently be modifed.)

	let sourceIsArray = Array.isArray(sourceIterable);
	let realization = sourceIsArray ? sourceIterable : [];
	let iterator = sourceIsArray ? null : undefined;

	function hasSourceIteratorNotBeenTouched() {
		return iterator === undefined;
	}

	function hasSourceIteratorBeenExhausted() {
		return iterator === null;
	}

	function markIteratorAsExhausted() {
		iterator = null;
	}

	function isResultIndexCached(index) {
		return index < realization.length;
	}

	// Used once to pull and cache results from the source iterable, one result at a time.

	function* iterateOverSource() {
		for (let result of asIterable(sourceIterable)) {
			realization.push(result); 
			yield result;
		}
	}

	// Exposed externally to support iteration. Will iterate over any cached results we hold before falling back on pulling from the source iterator.

	function* iterateOverRealization() {
		let i = -1;
		while (true) {
			++i;
			if (isResultIndexCached(i)) {
				yield realization[i];
				continue;
			}
			if (hasSourceIteratorNotBeenTouched()) {
				iterator = iterateOverSource();
			}
			if (hasSourceIteratorBeenExhausted()) {
				break;
			}
			let step = iterator.next();
			if (step.done) {
				markIteratorAsExhausted();
				break;
			}
			yield realization[i] = step.value;
		}
	}

	// Allows access to results by index. Cached results will be used if found; if not, the source iterator will be rolled forward just far enough to satisfy the request.

	function getAtIndex(index) {
		if (!isResultIndexCached(index)) {
			if (hasSourceIteratorNotBeenTouched()) {
				iterator = iterateOverSource();
			}
			if (hasSourceIteratorBeenExhausted()) {
				return;
			}
			while (!isResultIndexCached(index)) {
				let step = iterator.next();
				if (step.done) {
					markIteratorAsExhausted();
					return;
				}
			}
		}
		return realization[index];
	}

	Object.defineProperty(getAtIndex, 'length', {
		get: () => {
			if (hasSourceIteratorNotBeenTouched()) {
				iterator = iterateOverSource();
			}			
			while (!hasSourceIteratorBeenExhausted()) {
				let step = iterator.next();
				if (step.done) {
					markIteratorAsExhausted();
				}
			}
			return realization.length;
		}
	});

	getAtIndex[Symbol.iterator] = iterateOverRealization;
	getAtIndex._memoization = getAtIndex;		// Don't let what we return be memoized again.
	return getAtIndex;
}


export {
	asIndexable,
	asIterable,
	asReiterable,
	map,
	memoize,
	reduce,
	toArray
};