'use strict';

/*
 * Functions to support manipulation of 2D Euclidian geometries.
 * 
 */

import { assert } from './assertions.js';
import { isRealNumber } from './number.js';
import { hasProperties } from './object.js';
import { asIndexable, asReiterable, map, memoize  } from './functional.js';
import { range, sum } from './math.js';

const zeroedVertex = vertexFromXY(0, 0);

function assertIsCoordinate(value) {
	assert(value, function isCoordinate(value) {
		return isRealNumber(value);
	});
}

function assertIsScalar(value) {
	assert(value, function isScalar(value) {
		return isRealNumber(value);
	});
}

function assertIsVector(value) {
	assert(value, function isVector(value) {
		return hasProperties(value, [ 'dx', 'dy' ]);		// All (x,y) tuples have different properties so that we can detect when we are, for example, misusing a vector as a vertex.
	});
}

function assertIsVertex(value) {
	assert(value, function isVertex(value) {
		return hasProperties(value, [ 'vx', 'vy' ]);
	});
}

function vectorFromXY(x, y) {
	assertIsCoordinate(x);
	assertIsCoordinate(y);
	return { dx: x, dy: y };
}

function vertexFromXY(x, y) {
	assertIsCoordinate(x);
	assertIsCoordinate(y);
	return { vx: x, vy: y };
}

// Create a unit-scale regular polygon (e.g. triangle, square, hexagon) for use in a tessellation. Returns a set of vectors that trace out the sides of the shape.

function regularPolygonVectorsFromEdgeCount(edgeCount) {
	function* generator() {
		for (let i = 0; i < edgeCount; ++i) {
			let radians = i * 2 * Math.PI / edgeCount;
			let edge = vectorFromXY(Math.cos(radians), Math.sin(radians));
			yield edge;
		}
	};
	return memoize(generator());		// If we don't memorize / make reiterable, we have to be really careful whether or not our caller ever passes our result down to something that calls us twice. See the notes on iterators in functional.js.
}

// Convert a path of vectors to a set of vertices. (e.g. The vectors (0,1), (1,0), (-1,0), (0,-1) are chained to produce the vertices (0,0), (1,0), (1,1), (0,1).) 

function verticesFromVectors(vectors) {
	let lastVertex = zeroedVertex;
	return map(vectors, vector => {
		assertIsVector(vector);
		let currentVertex = lastVertex;
		lastVertex = vertexTranslatedByVector(lastVertex, vector);
		return currentVertex;
	});
}

// Move a vertex by some distance.

function vertexTranslatedByVector(vertex, vector) {
	assertIsVertex(vertex);
	assertIsVector(vector);
	return vertexFromXY(vertex.vx + vector.dx, vertex.vy + vector.dy);
}

// Scale a vector.

function vectorMultipliedByScalar(vector, scalar) {
	assertIsVector(vector);
	assertIsScalar(scalar);
	return vectorFromXY(vector.dx * scalar, vector.dy * scalar);
}

// Calculate the vector that would translate one vertex to another.

function vectorFromVertexToVertex(fromVertex, toVertex) {
	assertIsVertex(fromVertex);
	assertIsVertex(toVertex);
	return vectorFromXY(toVertex.vx - fromVertex.vx, toVertex.vy - fromVertex.vy)
}

// Calculate the halfway point between one vertex and another. 

function midpointOfVertices(fromVertex, toVertex) {
	let edgeVector = vectorFromVertexToVertex(fromVertex, toVertex);
	let midpointVector = vectorMultipliedByScalar(edgeVector, 0.5);
	let midpointVertex = vertexTranslatedByVector(fromVertex, midpointVector);
	return midpointVertex;
}

// Calculate all the midpoints for all edges of a shape.

function edgeMidpointVerticesOfPolygon(polygon) {
	if (polygon._edgeMidpointVertices) {
		return polygon._edgeMidpointVertices;
	}
	function* generator() {
		let firstVertex = null;
		let previousVertex = null;
		for (let currentVertex of polygon) {
			if (!firstVertex) {
				firstVertex = currentVertex;
			}
			if (previousVertex) {
				yield midpointOfVertices(previousVertex, currentVertex);
			}
			previousVertex = currentVertex;
		}
		yield midpointOfVertices(previousVertex, firstVertex);
	}
	return polygon._edgeMidpointVertices = asReiterable(generator);
}

// The angle, in radians, between the x-axis and the given vector.

function counterClockwiseAngleAtXAxisOfVector(vector) {
	assertIsVector(vector);
	return Math.atan2(vector.dy, vector.dx);
}

// The angle, in radians, between the first vector and the second.

function angleBetweenVectors(fromVector, toVector) {
	assertIsVector(fromVector);
	assertIsVector(toVector);
	let fromTheta = counterClockwiseAngleAtXAxisOfVector(fromVector);
	let toTheta = counterClockwiseAngleAtXAxisOfVector(toVector);
	return toTheta - fromTheta;
}

// Rotate the vector by the given angle.

function rotate(vector, radians) {
	let ca = Math.cos(radians);
	let sa = Math.sin(radians);
	return vectorFromXY(vector.dx * ca - vector.dy * sa, vector.dx * sa + vector.dy * ca);
}

// Calculate the centre point of an n-sided polygon.

function centroidOfPolygon(polygon) {
	if (polygon._centroid) {
		return polygon._centroid;		// A caching mechanism so that we can call this function repeatedly without worrying too much about performance.
	}

	let vertices = asIndexable(polygon);
	let n = vertices.length;

	function vertex(i) {
		return vertices(i % n);
	}

	function x(i) {
		return vertex(i).vx;
	}

	function y(i) {
		return vertex(i).vy;
	}

	// https://en.wikipedia.org/wiki/Centroid 

	const sumOverVertices = f => sum(map(range(0, n - 1), f));

	let a = (1 / 2) * sumOverVertices(i => x(i) * y(i + 1) - x(i + 1) * y(i));
	let cx = 1 / (6 * a) * sumOverVertices(i => (x(i) + x(i + 1)) * (x(i) * y(i + 1) - x(i + 1) * y(i)));
	let cy = 1 / (6 * a) * sumOverVertices(i => (y(i) + y(i + 1)) * (x(i) * y(i + 1) - x(i + 1) * y(i)));

	return polygon._centroid = vertexFromXY(cx, cy);
}

export {
	angleBetweenVectors,
	assertIsVector,
	assertIsVertex,
	centroidOfPolygon,
	counterClockwiseAngleAtXAxisOfVector,
	edgeMidpointVerticesOfPolygon,
	midpointOfVertices,
	regularPolygonVectorsFromEdgeCount,
	rotate,
	vectorFromXY,
	vectorFromVertexToVertex,
	vectorMultipliedByScalar,
	vertexFromXY,
	vertexTranslatedByVector,
	verticesFromVectors,
	zeroedVertex
};