'use strict';

/*
 * Generate a maze.
 * 
 */

import { range, round } from './math.js';
import { asIndexable, toArray, memoize, reduce } from './functional.js';
import { exitVectorsFromTile } from './tessellation.js';
import { rotate, vectorMultipliedByScalar, assertIsVector, assertIsVertex, add, centroidOfPolygon, vector, vertex, verticesFromVectors, vertexFromXY, zeroedVertex, vertexTranslatedByVector, vectorFromXY } from './geometry.js';

function randomIndexFromCount(count) {
	return Math.floor(Math.random() * count);
}

function roundVertex(vertex) {
	return vertexFromXY(round(vertex.vx, 1), round(vertex.vy, 1));
}

function putItemInCacheUnderKey(item, cache, key) {
	return cache[key] = item;
}

function itemFromCacheUnderKey(cache, key) {
	return cache[key];
}

// This is not particularly functional, but it's an approach I'm familiar with, and I can't think of a good immutable, functional implementation.

function putTileInCache(tile, cache) {
	let centroid = centroidOfPolygon(tile);
	let xDimensionKey = round(centroid.vx,1);
	let yDimensionKey = round(centroid.vy,1);
	let yDimensionCache = itemFromCacheUnderKey(cache, xDimensionKey);
	if (!yDimensionCache) {
		yDimensionCache = putItemInCacheUnderKey({}, cache, xDimensionKey);
	}
	let cachedTile = putItemInCacheUnderKey(tile, yDimensionCache, yDimensionKey);
	return cachedTile;
}

function isTileInCache(tile, cache) {
	let centroid = centroidOfPolygon(tile);
	return !!tileFromCentroidOutOfCache(centroid, cache);
}

function tileFromCentroidOutOfCache(centroid, cache) {
	let xDimensionKey = round(centroid.vx,1);
	let yDimensionKey = round(centroid.vy,1);
	let yDimensionCache = itemFromCacheUnderKey(cache, xDimensionKey);
	if (yDimensionCache) {
		return yDimensionCache[yDimensionKey];
	}
}

// The basic approach is to start with a tessellation of tiles and treat each edge as a potential bridge to another tile. This means we can have interesting mazes rather than
// just square ones.

function generateRandomPathInTessellationFromTileAvoidingTiles(tessellation, fromTile, avoidingTiles, isTileAcceptable) {
	function* generator() {
		let currentTile = asIndexable(fromTile);
		let centroid = centroidOfPolygon(currentTile);
		yield centroid;
		putTileInCache(currentTile, avoidingTiles);
		while (true) {
			let possibleExitIndexes = toArray(range(0, currentTile.length - 1));
			while (true) {
				if (!possibleExitIndexes.length) {
					return;
				}
				let possibleExitIndex = randomIndexFromCount(possibleExitIndexes.length);
				let chosenExitIndex = possibleExitIndexes[possibleExitIndex];
				let nextTile = tessellation.getNeigbourOfTileForExitIndex(currentTile, chosenExitIndex);
				let nextTileCentroid = centroidOfPolygon(nextTile);
				let tileHasAlreadyBeenVisited = isTileInCache(nextTile, avoidingTiles);
				if (!tileHasAlreadyBeenVisited && isTileAcceptable(nextTile)) {
					putTileInCache(nextTile, avoidingTiles);
					yield nextTileCentroid;
					currentTile = nextTile;
					break;
				}
				possibleExitIndexes.splice(possibleExitIndex, 1);
			}
		}
	}
	return generator();
}

function mazeFromTessellationWithinBoundingVector(tessellation, boundingVector) {
	let startingShape = asIndexable(tessellation.startingShape);
	let startingTile = verticesFromVectors(startingShape);
	let unitVector = startingShape(0);

	// Floating-point inaccuracies can mess with the cache hit tests. We round vertices to avoid this. The rounding means we need a bit of leeway when testing whether
	// tiles are within bounds.

	let leftToRightMarginOfErrorVector = vectorMultipliedByScalar(unitVector, 0.5);		 
	let topRightMarginOfErrorVector = rotate(leftToRightMarginOfErrorVector, ((2 * Math.PI) / 8) * 1);
	let bottomLeftMarginOfErrorVector = rotate(leftToRightMarginOfErrorVector, ((2 * Math.PI) / 8) * 5);
	let bottomLeftBound = centroidOfPolygon(startingTile);
	let topRightBound = vertexTranslatedByVector(bottomLeftBound, boundingVector);
	let avoidingTiles = {};

	bottomLeftBound = vertexTranslatedByVector(bottomLeftBound, bottomLeftMarginOfErrorVector);
	topRightBound = vertexTranslatedByVector(topRightBound, topRightMarginOfErrorVector);

	function isTileWithinBounds(tile) {
		let centroid = centroidOfPolygon(tile);
		return centroid.vx >= bottomLeftBound.vx && centroid.vy >= bottomLeftBound.vy && centroid.vx <= topRightBound.vx && centroid.vy <= topRightBound.vy;	
	}

	// Seed the maze with a random path.

	let path = generateRandomPathInTessellationFromTileAvoidingTiles(tessellation, startingTile, avoidingTiles, isTileWithinBounds);
	let seed = memoize(reduce(path, (accumulator, vertex) => (accumulator.push(vertex), accumulator), []));
	let lines = [ seed ];

	// Fill in the gaps, ensuring not to walk into a tile we've already visited.

	for (let i = 0; i < lines.length; ++i) {
		let line = lines[i];
		for (let centroid of line) {
			let tile = tileFromCentroidOutOfCache(centroid, avoidingTiles);
			let path = generateRandomPathInTessellationFromTileAvoidingTiles(tessellation, tile, avoidingTiles, isTileWithinBounds);
			let line = memoize(reduce(path, (accumulator, vertex) => (accumulator.push(vertex), accumulator), []));
			if (line.length > 1) {		// If we couldn't expand the maze, we get just the starting vertex back.
				lines.push(line);
			}
		}
	}

	return {
		lines: lines 
	};
}

export { mazeFromTessellationWithinBoundingVector };