'use strict';

/*
 * Render an SVG version of the maze.
 * 
 */

import { assert } from './assertions.js';
import { isRealNumber } from './number.js';
import { hasProperties } from './object.js';
import { map, reduce } from './functional.js';
import { join, joinSeparatedBySpaces } from './string.js';
import { assertIsVertex } from './geometry.js';
import { round } from './math.js';

const zeroedBounds = boundsFromTopLeftBottomRight(0, 0, 0, 0);

function assertIsPoint(value) {
	assert(value, function isPoint(value) {
		return hasProperties(value, [ 'px', 'py' ]);
	});
}

function assertIsBound(value) {
	assert(value, function isBound(value) {
		return isRealNumber(value);
	});
}

function assertIsBounds(value) {
	assert(value, function isBound(value) {
		return hasProperties(value, [ 'top', 'left', 'bottom', 'right' ]);
	});
}

function assertIsCoordinate(value) {
	assert(value, function isCoordinate(value) {
		return isRealNumber(value);
	});
}

function serializeAttributes(attributes) {
	return reduce(attributes, joinSeparatedBySpaces(attribute => `${attribute.name}="${attribute.value}"`)); // TODO escaping
}

function serializeBounds(bounds) {
	assertIsBounds(bounds);
	return `${round(bounds.left)} ${round(bounds.top)} ${round(bounds.right - bounds.left)} ${round(bounds.bottom - bounds.top)}`;
}

function tagFrom(name, attributes, content) {
	let attributeString = serializeAttributes(attributes);
	if (attributeString) {
		attributeString = ' ' + attributeString;
	}
	return content ? `<${name}${attributeString}>${content}</${name}>` : `<${name}${attributeString}/>`;
}

function pointFromXY(x, y) {
	assertIsCoordinate(x);
	assertIsCoordinate(y);
	return {
		px: x,
		py: y
	};
}

function pointFromVertex(vertex) {
	assertIsVertex(vertex);
	return pointFromXY(vertex.vx, -vertex.vy);
}

function serializePoint(point) {
	assertIsPoint(point);
	return `${round(point.px)},${round(point.py)}`;
}

function serializePoints(points) {
	return reduce(points, joinSeparatedBySpaces(point => serializePoint(point)));
}

function boundsFromTopLeftBottomRight(top, left, bottom, right) {
	assertIsBound(top);
	assertIsBound(left);
	assertIsBound(bottom);
	assertIsBound(right);
	return {
		top: top,
		left: left,
		bottom: bottom,
		right: right
	};
}

function boundsFromPoints(points) {
	return reduce(
		points,
		(accumulator, point) => boundsFromTopLeftBottomRight(
			Math.min(accumulator.top, point.py),
			Math.min(accumulator.left, point.px),
			Math.max(accumulator.bottom, point.py),
			Math.max(accumulator.right, point.px)
		),
		zeroedBounds);
}

function pointsRenderingFromVertices(tagName, vertices, attributes) {
	let points = map(vertices, vertex => pointFromVertex(vertex));
	let result = {
		bounds: boundsFromPoints(points),
		markup: tagFrom(tagName, reduce(attributes, (accumulator, attribute) => (accumulator[attribute.name] = attribute.value, accumulator), { 'points': serializePoints(points) }))
	};
	return result;
}

function drawPolygon(vertices) {
	return pointsRenderingFromVertices('polygon', vertices);
}

function drawPolyline(vertices) {
	return pointsRenderingFromVertices('polyline', vertices, { fill: 'none' });
}

function render(renderers) {
	let viewBounds = zeroedBounds;
	let markup = reduce(renderers, join(renderer => {
		let rendering = renderer();
		viewBounds = boundsFromTopLeftBottomRight(
			Math.min(viewBounds.top, rendering.bounds.top),
			Math.min(viewBounds.left, rendering.bounds.left),
			Math.max(viewBounds.bottom, rendering.bounds.bottom),
			Math.max(viewBounds.right, rendering.bounds.right)
		);
		return rendering.markup;
	}));
	let margin = 0.5;
	viewBounds = boundsFromTopLeftBottomRight(round(viewBounds.top - margin), round(viewBounds.left - margin), round(viewBounds.bottom + margin), round(viewBounds.right + margin));
	return tagFrom('svg', { viewBox: serializeBounds(viewBounds), preserveAspectRatio: 'xMidYMid meet' }, markup);
}

export {
	drawPolygon
,	drawPolyline
,	render
};