'use strict';

/*
 * Cover a plane with tiles.
 * 
 */

import { round } from './math.js';
import { map, memoize, asIndexable } from './functional.js';
import { add, angleBetweenVectors, centroidOfPolygon, edgeMidpointVerticesOfPolygon, midpointOfVertices, vectorFromVertexToVertex, counterClockwiseAngleAtXAxisOfVector, rotate, verticesFromVectors, regularPolygonVectorsFromEdgeCount, vertexTranslatedByVector, vertexFromXY } from './geometry.js';

function exitVectorsFromTile(tile) {
	let centreVertex = centroidOfPolygon(tile);
	let midpointVertices = edgeMidpointVerticesOfPolygon(tile);
	return map(midpointVertices, midpointVertex => vectorFromVertexToVertex(centreVertex, midpointVertex));
}

const triangle = regularPolygonVectorsFromEdgeCount(3);
const square = regularPolygonVectorsFromEdgeCount(4);

function getNeighbourOfTileInShape(tile, exitIndex, f) {
	let exitVectors = asIndexable(exitVectorsFromTile(tile));
	let exitVector = exitVectors(exitIndex);

	let neighbouringShape = memoize(f());

	let edgeMidpointVertices = memoize(edgeMidpointVerticesOfPolygon(tile));
	let exitEdgeMidpointVertex = edgeMidpointVertices(exitIndex); 

	let neighbouringAlignmentVector = rotate(neighbouringShape(0), Math.PI / 2); // Perpendicular to aligment edge, pointing inside the shape.
	let adjustmentAngle = angleBetweenVectors(neighbouringAlignmentVector, exitVector);
	let neighbouringVectors = map(neighbouringShape, vector => rotate(vector, adjustmentAngle));
	let neighbouringTile = memoize(verticesFromVectors(neighbouringVectors));
	let neighbouringEdgeMidpointVertex = midpointOfVertices(neighbouringTile(0), neighbouringTile(1));
	let requiredTranslation = vectorFromVertexToVertex(neighbouringEdgeMidpointVertex, exitEdgeMidpointVertex);
	return map(neighbouringTile, vertex => vertexTranslatedByVector(vertex, requiredTranslation));
}

const squareTessellation = {
	startingShape: square,
	getNeigbourOfTileForExitIndex(tile, exitIndex) {
		return getNeighbourOfTileInShape(tile, exitIndex, () => square);
	}
};

const triangleTessellation = {
	startingShape: triangle,
	getNeigbourOfTileForExitIndex(tile, exitIndex) {
		return getNeighbourOfTileInShape(tile, exitIndex, () => triangle);
	}
};

const squareTriangleTessellation = {
	startingShape: square,
	getNeigbourOfTileForExitIndex(tile, exitIndex) {
		let exitVectors = asIndexable(exitVectorsFromTile(tile));
		let exitVector = exitVectors(exitIndex);
		let exitVectorAngle = counterClockwiseAngleAtXAxisOfVector(exitVector);

		function getNeighbouringShape() {
			if (exitVectors.length == 4) {
				return triangle;
			}
			if (exitVectors.length == 3) {
				switch (round(exitVectorAngle, 1)) {
					case 1.6:
					case -0.5:
					case -3.1:
					case 1:
					case -1.6:
					case 0:
					case -2.1:
					case 3.1:
					case 2.6:
						return square;
					case -2.6:
					case -1:
					case 0.5:
					case 2.1:
						return triangle;
				}
			}
			throw Error('Cannot deduce neighbour. ' + exitVectorAngle + '; ' + round(exitVectorAngle, 5) + ' ' + round(exitVectorAngle, 1));
		}
		return getNeighbourOfTileInShape(tile, exitIndex, getNeighbouringShape);
	}
};

export {
	exitVectorsFromTile,
	squareTessellation,
	squareTriangleTessellation,
	triangleTessellation
};